/*  Routines to keep track of the HCI state and to update the LCD accordingly

*/

/*-----------------------------------------------------------------------------------------
   Structure and tables used to define the Menus
  ----------------------------------------------------------------------------------------*/

/*Field enunerations that provide a symbolic way of refering the selected field.  These
   enumerations allow using a Switch - case structure to determine what action needs to
   be taken for each field.  The HCI routine only keeps track of the page and field on
   that page, but there is a look-up initialized the HCI_init routine that converts the
   page/local field info to these enumerated values  */

enum field_ID
{volume, input, spkr_sel, spkr_offset, IR_action, IR_Key_map };

/*The constants that follow are the captions and display values for the LCD display
  They are stored in program space using PROGMEM so they don't waste RAM */

const byte vol_caption_len = 12;
const char vol_caption[vol_caption_len + 1] PROGMEM = {"Volume in dB"};
const byte vol_value_qty = 91;
const byte vol_value_max = 10;
const byte vol_value_len = 3;
const char vol_values[vol_value_qty][vol_value_len + 1] PROGMEM = {
  {"-80"}, {"-79"}, {"-78"}, {"-77"}, {"-76"}, {"-75"}, {"-74"}, {"-73"}, {"-72"}, {"-71"},
  {"-70"}, {"-69"}, {"-68"}, {"-67"}, {"-66"}, {"-65"}, {"-64"}, {"-63"}, {"-62"}, {"-61"},
  {"-60"}, {"-59"}, {"-58"}, {"-57"}, {"-56"}, {"-55"}, {"-54"}, {"-53"}, {"-52"}, {"-51"},
  {"-50"}, {"-49"}, {"-48"}, {"-47"}, {"-46"}, {"-45"}, {"-44"}, {"-43"}, {"-42"}, {"-41"},
  {"-40"}, {"-39"}, {"-38"}, {"-37"}, {"-36"}, {"-35"}, {"-34"}, {"-33"}, {"-32"}, {"-31"},
  {"-30"}, {"-29"}, {"-28"}, {"-27"}, {"-26"}, {"-25"}, {"-24"}, {"-23"}, {"-22"}, {"-21"},
  {"-20"}, {"-19"}, {"-18"}, {"-17"}, {"-16"}, {"-15"}, {"-14"}, {"-13"}, {"-12"}, {"-11"},
  {"-10"}, {"-9 "}, {"-8 "}, {"-7 "}, {"-6 "}, {"-5 "}, {"-4 "}, {"-3 "}, {"-2 "}, {"-1 "},
  {" 0 "}, {"+1 "}, {"+2 "}, {"+3 "}, {"+4 "}, {"+5 "}, {"+6 "}, {"+7 "}, {"+8 "}, {"+9 "},
  {"+10"}
};

//Note:  the string lengths are increased by 1 in the array declarations to allow space for the null
const byte input_caption_len = 5;
const char input_caption[input_caption_len + 1] PROGMEM = {"Input"};
const byte input_value_qty = 3;
const byte input_value_len = 6;
const char input_values[input_value_qty][input_value_len + 1] PROGMEM = {{"  CD  "}, {"  PC  "}, {"  DVD "}};


const byte chan_vol_caption_len = 16;
const char chan_vol_caption[chan_vol_caption_len + 1] PROGMEM = {"Speaker   Offset"};
const byte chan_sel_qty = 5;
const byte chan_sel_len = 3;
const char chan_sel_values[chan_sel_qty][chan_sel_len + 1] PROGMEM = {{"Off"}, {"  1"}, {"  2"}, {"  3"}, {"  4"}};
const byte chan_trim_qty = 25;

//Calculate the offset needed to get 0db from the PGA2311:
const byte PGA2311_offset = 192 + (vol_value_max * 2) - ((vol_value_qty - 1) * 2) - chan_trim_qty / 2;

const byte chan_trim_len = 4;
const char chan_trim_values[chan_trim_qty][chan_trim_len + 1] PROGMEM = {
  {"-6.0"}, {"-5.5"}, {"-5.0"}, {"-4.5"}, {"-4.0"}, {"-3.5"}, {"-3.0"}, {"-2.5"}, {"-2.0"}, {"-1.5"}, {"-1.0"}, {"-0.5"}, {" 0  "},
  {"+0.5"}, {"+1.0"}, {"+1.5"}, {"+2.0"}, {"+2.5"}, {"+3.0"}, {"+3.5"}, {"+4.0"}, {"+4.5"}, {"+5.0"}, {"+5.5"}, {"+6.0"}
};


const byte IR_caption_len = 14;
const char IR_caption[IR_caption_len + 1] PROGMEM = {"IR Remote Cont"};
const byte IR_action_qty = 10;
const byte IR_action_len = 5;

enum Remote_key {left, right, up, down, zero, one, two, three, four, mute};   //this must match the labels in the next line

const char IR_action_values[IR_action_qty][IR_action_len + 1] PROGMEM = {{" left"}, {"right"}, {"  up "}, {" down"}, {"  0  "}, {"  1  "}, {"  2  "}, {"  3  "}, {"  4  "}, {" Mute"}};
const byte IR_key_qty = 1;
const byte IR_key_len = 3;
const char IR_key_values[IR_key_qty][IR_key_len + 1] PROGMEM = {{"<?>"}};



/*The HCI_init routine sets up the pages by specifying the caption and the fields on each page.
   The captions are the static text data, and the fields have the variable data.  The field data
   can range from 0 to number_of_values.  What gets displayed in the field is a looked up value
   from the tables in Program memory, using the field value as an index.

   The row and column numbers are the locations for the cursor to start printing to the LCD.
   Each field can have a caption, but if there is no caption for that field, the length is 0.
*/

void  HCI_init() {
  HCI.number_of_pages = number_of_menu_pages;
  //Page 0 = Volume
  HCI.menupage[0].number_of_fields = 1;
  HCI.menupage[0].page_field[0].caption_length = vol_caption_len;
  HCI.menupage[0].page_field[0].caption_table_pointer = &vol_caption[0];
  HCI.menupage[0].page_field[0].caption_row = 0;
  HCI.menupage[0].page_field[0].caption_column = 2;
  HCI.menupage[0].page_field[0].number_of_values = vol_value_qty;
  HCI.menupage[0].page_field[0].value_length = vol_value_len;
  HCI.menupage[0].page_field[0].value_table_pointer = &vol_values[0][0];
  HCI.menupage[0].page_field[0].value_row = 1;
  HCI.menupage[0].page_field[0].value_column = 6;
  HCI.menupage[0].page_field[0].field_value = 50;     //default value for volume = -40
  HCI.menupage[0].page_field[0].field_ID = volume;

  //Page 1 = Input
  HCI.menupage[1].number_of_fields = 1;
  HCI.menupage[1].page_field[0].caption_length = input_caption_len;
  HCI.menupage[1].page_field[0].caption_table_pointer = &input_caption[0];
  HCI.menupage[1].page_field[0].caption_row = 0;
  HCI.menupage[1].page_field[0].caption_column = 5;
  HCI.menupage[1].page_field[0].number_of_values = input_value_qty;
  HCI.menupage[1].page_field[0].value_length = input_value_len;
  HCI.menupage[1].page_field[0].value_table_pointer = &input_values[0][0];
  HCI.menupage[1].page_field[0].value_row = 1;
  HCI.menupage[1].page_field[0].value_column = 5;
  HCI.menupage[1].page_field[0].field_value = 0;      //default value = Analog
  HCI.menupage[1].page_field[0].field_ID = input;

  //Page 2 = Relay/Volume Offset
  HCI.menupage[2].number_of_fields = 2;
  HCI.menupage[2].page_field[0].caption_length = chan_vol_caption_len;
  HCI.menupage[2].page_field[0].caption_table_pointer = &chan_vol_caption[0];
  HCI.menupage[2].page_field[0].caption_row = 0;
  HCI.menupage[2].page_field[0].caption_column = 0;
  HCI.menupage[2].page_field[0].number_of_values = chan_sel_qty;
  HCI.menupage[2].page_field[0].value_length = chan_sel_len;
  HCI.menupage[2].page_field[0].value_table_pointer = &chan_sel_values[0][0];
  HCI.menupage[2].page_field[0].value_row = 1;
  HCI.menupage[2].page_field[0].value_column = 1;
  HCI.menupage[2].page_field[0].field_value = 0;     //default trim chan = 1
  HCI.menupage[2].page_field[0].field_ID = spkr_sel;

  HCI.menupage[2].page_field[1].caption_length = chan_vol_caption_len;
  HCI.menupage[2].page_field[1].caption_table_pointer = &chan_vol_caption[0];
  HCI.menupage[2].page_field[1].caption_row = 0;
  HCI.menupage[2].page_field[1].caption_column = 0;
  HCI.menupage[2].page_field[1].number_of_values = chan_trim_qty;
  HCI.menupage[2].page_field[1].value_length = chan_trim_len;
  HCI.menupage[2].page_field[1].value_table_pointer = &chan_trim_values[0][0];
  HCI.menupage[2].page_field[1].value_row = 1;
  HCI.menupage[2].page_field[1].value_column = 10;
  HCI.menupage[2].page_field[1].field_value = 12;     //default trim value = 0
  HCI.menupage[2].page_field[1].field_ID = spkr_offset;


  //Page 3 = IR Remote Cont
  HCI.menupage[3].number_of_fields = 2;
  HCI.menupage[3].page_field[0].caption_length = IR_caption_len;
  HCI.menupage[3].page_field[0].caption_table_pointer = &IR_caption[0];
  HCI.menupage[3].page_field[0].caption_row = 0;
  HCI.menupage[3].page_field[0].caption_column = 1;
  HCI.menupage[3].page_field[0].number_of_values = IR_action_qty;
  HCI.menupage[3].page_field[0].value_length = IR_action_len;
  HCI.menupage[3].page_field[0].value_table_pointer = &IR_action_values[0][0];
  HCI.menupage[3].page_field[0].value_row = 1;
  HCI.menupage[3].page_field[0].value_column = 1;
  HCI.menupage[3].page_field[0].field_value = 0;     //default IR action = select
  HCI.menupage[3].page_field[0].field_ID = IR_action;

  HCI.menupage[3].page_field[1].caption_length = IR_caption_len;
  HCI.menupage[3].page_field[1].caption_table_pointer = &IR_caption[0];
  HCI.menupage[3].page_field[1].caption_row = 0;
  HCI.menupage[3].page_field[1].caption_column = 1;
  HCI.menupage[3].page_field[1].number_of_values = IR_key_qty;
  HCI.menupage[3].page_field[1].value_length = IR_key_len;
  HCI.menupage[3].page_field[1].value_table_pointer = &IR_key_values[0][0];
  HCI.menupage[3].page_field[1].value_row = 1;
  HCI.menupage[3].page_field[1].value_column = 10;
  HCI.menupage[3].page_field[1].field_value = 0;
  HCI.menupage[3].page_field[1].field_ID = IR_Key_map;


  current_page = 0;
  current_page_field = 0;
  current_field_ID = 0;
}


void  LCD_init() {
  // set up the LCD's number of columns and rows:
  lcd.begin(16, 2);
  // Print a message to the LCD.
  lcd.setCursor(0, 0);
  lcd.print("Ready");
  lcd.setCursor(0, 1);
}

void next_field() {
  //Update the field pointer, then check if need to go to the next page
  current_page_field++;
  if (current_page_field == HCI.menupage[current_page].number_of_fields) {
    if (current_page != HCI.number_of_pages-1) {
      current_page++;
      current_page_field = 0;
    }
    else {
      current_page = 0;
      current_page_field = 0;
    }
  }

  //If past end of pages, reset page pointer to zero
  if ((current_page > HCI.number_of_pages) && (current_page_field == HCI.menupage[current_page].number_of_fields)) {
    current_page = 0;
    current_page_field = 0;
  }
  HCI_page_refresh();
}

void previous_field() {
  if (current_page_field > 0) {
    current_page_field--;
    HCI_page_refresh();
  }
  else {
    if (current_page == 0) {
      current_page =  HCI.number_of_pages - 1;
      current_page_field = 0;
      HCI_page_refresh();
    }
    else {
      current_page--;
      current_page_field = 0;
      HCI_page_refresh();
    }
  }
}


void previous_value() {
  if (HCI.menupage[current_page].page_field[current_page_field].field_value > 0 ) {
    HCI.menupage[current_page].page_field[current_page_field].field_value--;
    print_field_values(current_page, current_page_field);
  }
  else  {
    HCI.menupage[current_page].page_field[current_page_field].field_value = HCI.menupage[current_page].page_field[current_page_field].number_of_values - 1;
    print_field_values(current_page, current_page_field);
  }
  update_from_HCI(current_page, current_page_field);
  save_state();
}


void next_value() {
  if (HCI.menupage[current_page].page_field[current_page_field].field_value < (HCI.menupage[current_page].page_field[current_page_field].number_of_values - 1) ) {
    HCI.menupage[current_page].page_field[current_page_field].field_value++;
    print_field_values(current_page, current_page_field);
  }
  else  {
    HCI.menupage[current_page].page_field[current_page_field].field_value = 0;
    print_field_values(current_page, current_page_field);
  }
  update_from_HCI(current_page, current_page_field);
  save_state();
}


void current_value(byte field_ID, byte field_value, bool update_display) {    //get here from the remote control number keys
  //this routine finds the page and field on that page for the field_ID variable.  Then it updates the value from field_value
  for (byte i = 0; i < HCI.number_of_pages; i++) {
    for (byte j = 0; j < HCI.menupage[i].number_of_fields; j++) {
      if (HCI.menupage[i].page_field[j].field_ID == field_ID) {
        HCI.menupage[i].page_field[j].field_value = field_value;
        update_from_HCI(i, j);
        if (update_display == true) {
          current_page = i;  current_page_field = j;
          HCI_page_refresh();
        }
      }
    }
  }

}


void print_caption(byte p, byte f) {
  byte len;
  int i = 0;
  byte r;
  byte c;
  char caption_string[17];

  len = HCI.menupage[p].page_field[f].caption_length;
  r = HCI.menupage[p].page_field[f].caption_row;
  c = HCI.menupage[p].page_field[f].caption_column;

  if (len > 0) {
    lcd.setCursor(c, r);
    for (i = 0; i < len; i++) {
      caption_string[i] = pgm_read_byte_near(HCI.menupage[p].page_field[f].caption_table_pointer + i);
    }
    caption_string[i] = '\0';
    lcd.print(caption_string);
  }
}

void print_field_values(byte p, byte f)  {
  byte len;
  int i;
  byte field_count;
  byte r;
  byte c;
  char field_string[17];
  int table_offset;
  int EEPROM_word;

  lcd.setCursor(0, 1);
  lcd.print("                ");    //clear out second row

  field_count = 0;
  while (field_count < HCI.menupage[p].number_of_fields)
  {
    len = HCI.menupage[p].page_field[field_count].value_length;
    r = HCI.menupage[p].page_field[field_count].value_row;
    c = HCI.menupage[p].page_field[field_count].value_column;

    lcd.setCursor(c, r);
    for (i = 0; i < len; i++) {
      table_offset = HCI.menupage[p].page_field[field_count].field_value * (len + 1);
      field_string[i] = pgm_read_byte_near(HCI.menupage[p].page_field[field_count].value_table_pointer + i + table_offset);
    }
    field_string[i] = '\0';

    //Exception for the IR remote field
    if (HCI.menupage[p].page_field[field_count].field_ID == IR_Key_map) {
      table_offset = IR_map_start + 2 * HCI.menupage[p].page_field[0].field_value;   //use IR action field value to find offset
      EEPROM_word = EEPROM.read(table_offset) * 256 + EEPROM.read(table_offset + 1);
      itoa(EEPROM_word, field_string, 16);  //(integer, Buffer, base)
    }
    //Continue from exception

    lcd.print(field_string);
    if ((f == field_count) && (HCI.menupage[p].number_of_fields > 1)) {
      lcd.print("\x7f");      //print the cursor (left arrow)
    }
    field_count++;
  }
}

void HCI_page_refresh() {
  lcd.clear();
  print_caption(current_page, current_page_field);
  print_field_values(current_page, current_page_field);
}



